# cody.py (Explorer Workshop)
# Copyright: 2024 Bradan Lane STUDIO
#
# Open Source License: MIT

'''
This CircuitPython program is a sample from the 'Hack the Badge' workshop

Implementation Notes
--------------------
**Hardware:**
    * one of the Explorer Badge designs

**Software and Dependencies:**
    * CircuitPython firmware (UF2): https://circuitpython.org
    * Adafruit CircuitPython Libraries: https://circuitpython.org/libraries

**Open Source Repository:**
    * GitLab: https://gitlab.com/bradanlane/exp_tutorial
    
'''

import board
import time
import digitalio
import audiopwmio
import synthio
import ulab.numpy as np # provides lots of numeric data processing


sound_enable = digitalio.DigitalInOut(board.SPEAKER_EN)
sound_enable.direction = digitalio.Direction.OUTPUT
sound_enable.value = False # we start with the amplifier OFF

# create a basic synth
SAMPLE_SIZE = 1024
SAMPLE_VOLUME = 32767 # largest positive number in a 16bit integer

wave = np.linspace(SAMPLE_VOLUME, -SAMPLE_VOLUME, num=SAMPLE_SIZE, dtype=np.int16)

synth = synthio.Synthesizer(sample_rate=22050)
audio = audiopwmio.PWMAudioOut(board.SPEAKER)
audio.play(synth)

# because we can have multiple notes at once, we need to track them
# some data to track notes and control the amplifier
notes = {} # dict used to track notes
note_count = 0

def tone_on(frequency:float=440.0):
    global wave, synth, sound_enable, notes, note_count
    ''' add a new note (frequency) to the polyphonic output '''
    # add to active notes
    if not note_count:
        sound_enable.value = True
    synth_note = synthio.Note(frequency, waveform=wave)
    # save the note so we can match it up when we stop playing it
    notes[frequency] = synth_note
    synth.press(synth_note)
    note_count += 1

def tone_off(frequency:float=440.0):
    global wave, synth, sound_enable, notes, note_count
    ''' remove an active note (frequency) from the polyphonic output '''
    synth_note = notes.get(frequency, None)
    if synth_note is not None:
        synth.release(synth_note)
        notes.pop(frequency)
        note_count -= 1
    if not note_count:
        sound_enable.value = False

def tone_off_all():
    global wave, synth, sound_enable, notes, note_count
    ''' remove all existing notes from the polyphonic output '''
    for note in notes:
        synth_note = notes.get(note, None)
        synth.release(synth_note)
        notes.pop(note)
    note_count = 0
    sound_enable.value = False

music_scale:list = [523.25, 587.33, 659.33, 698.46, 783.99, 880.00, 987.77, 1046.50, 1174.66]

# play each note - building quite a cacophony
for note in music_scale:
    tone_on(note)
    time.sleep(0.25)
tone_off_all()

counter:int = 0

try:
    while True:
        # stop prior chord
        tone_off_all()
        # play a chord
        for i in range(0, 8, 2):
            index = (counter + i) % len(music_scale)
            tone_on(music_scale[index])
            time.sleep(0.125)
        time.sleep(0.25)
        counter += 1

finally:
    # any cleanup code goes here
    pass
